(* Time-stamp: <modified the 05/09/2007 (at 09:57) by Erwan Jahier>  *)

(* 
   This is the XML Schema loader.
   It produces an OCaml structure from an XML tree. 

   
*)

{{ namespace core="file:///tmp/aadl-schema/core" }}
(* {{ namespace core="http:///www-verimag.imag.fr/~jahier/xsd/AADLXMI/core" }} *)

{{ namespace xsi="http://www.w3.org/2001/XMLSchema-instance"  }}

open Aadl_ast
open Aadl_dtd
open Ident

(* exception Cycle of {{ Atom }} *)
  (** Raised in presence of an ill-formed recursion in a .aaxl *)


(**********************************************************************************)
(* parsing xml using Xml ligth *)

let (from_xml : ?ns:bool -> Xml.xml -> Ocamlduce.Load.anyxml) = 
  fun ?ns xml -> 
    let is_not_a_comment = function
	| Xml.Element (tag, _ , _) -> tag <> "comment"
	| Xml.PCData _ -> true
    in
    let l = Ocamlduce.Load.make ?ns () in
    let rec aux = function
      | Xml.Element (tag, attrs, child0) ->
	  let child = List.filter is_not_a_comment child0 in
	    (
	      Ocamlduce.Load.start_elem l tag attrs; 
	      List.iter aux child; 
	      Ocamlduce.Load.end_elem l ()
	    )
      | Xml.PCData s ->
	  Ocamlduce.Load.text l s
    in
      aux xml;
      Ocamlduce.Load.get l

let (from_file : ?ns:bool -> string -> Ocamlduce.Load.anyxml) = 
  fun ?ns s ->  from_xml ?ns (Xml.parse_file s)


(**********************************************************************************)
open Aadl_dtd


(**
   Load the xml file and check it is well-typed according to the ocamlduce
   Type definition of [Aadl_dtd.ml]
*)
let (load_aaxl : int -> string -> Aadl_dtd.aadlSpec) =
  fun verb_level file_name -> 
    let load_xml = from_file ~ns:true in
    let aadl_spec =
      try {{ (load_xml file_name :? aadlSpec ) }}
      with 
	  Failure s ->
	    if verb_level >= 2 then
	      (
		Printf.eprintf "*** Invalid AAXL document:%s@." s;
		flush stdout
	      )
	    else
	      (
		print_string "*** Invalid AAXL document: it migth be due to: \n";
		print_string "*** \t (1) the header, which ougth to say where to find the aaxl schema, points to a wrong place. Have you set correctly set the AADL_SCHEMA_DIR environment variable?\n";
		print_string "*** \t (2) an aaxl unsupported entitier. Please send jahier@imag.fr to report that.\n";
		print_string "*** \t (3) a bug in the osate aaxl generator (which ougth to generate schema conforming code).\n";
		print_string "*** use -v 2 for more information on this schema error.\n")
	    ;
	    flush stdout;
	    exit 2
	      
	| Xml.Error(e) -> 
	    print_string ("*** XML parsing error if file " ^ file_name ^ 
			    ": " ^ (Xml.error e) ^ "\n");
	    flush stdout;
	    exit 2
	      	    
    in
(*       print_string ("The content of " ^ file_name ^ " seems to be ok\n"); *)
(*       flush stdout; *)
      aadl_spec

(**********************************************************************************)

(* I redefine my own string translator as Ocamlduce.to_string returns
   a string surrounded by double quotes (e.g., "\"toto\"", instead of
   just "toto").
*)
let (ocamlduce_to_string : {{ String }} -> string) =
  fun duce_str -> 
    let str' = Ocamlduce.to_string duce_str in
      String.sub str' 1 ((String.length str') - 2)


(**********************************************************************************)
let (string_to_ident : {{ String }} -> Ident.t) =
  fun str -> 
    Ident.aaxl_string_to_ident (ocamlduce_to_string str)

let (aaxl_string_to_ident : {{ String }} -> Ident.t) =
  fun str -> 
    Ident.aaxl_string_to_ident (ocamlduce_to_string str)


let (string_to_long_ident : {{ String }} -> Ident.t) =
  fun str -> 
    Ident.aaxl_string_to_long_ident (ocamlduce_to_string str)


let (access_connection_string_to_ident : string option -> {{  [String?] }} -> 
      {{ String }} -> Ident.t) =
  fun suff_opt str1 str2 -> 
    let id =
      match str1 with
	  {{ [x] }} -> 
	    Ident.connection_string_to_ident 
	      (ocamlduce_to_string x) (ocamlduce_to_string str2)
	| {{ [] }} -> aaxl_string_to_ident str2
    in
      match suff_opt, id with
	| None, _ -> id
	| Some suff, Ident _ -> id (* *)
	| Some suff, _ ->  Ident.add_suffix suff id
    


let (seq_string_to_ident_opt : {{ [String?] }} -> Ident.t option ) = function
    {{ [x] }} -> Some (string_to_ident x)
  | {{ [] }} -> None
      

let (seq_string_to_long_ident_opt : {{ [String?] }} -> Ident.t option ) = function
    {{ [x] }} -> Some (string_to_long_ident x)
  | {{ [] }} -> None


(**********************************************************************************)

let (aadl_ident_opt_to_bool : Ident.t option -> bool) = function
    None -> assert false
  | Some x  -> print_string (Ident.to_string x);assert false

let (aadl_ident_opt_to_addlint : Ident.t option -> Ident.t option -> 
      aadl_intger_or_constant) =
  fun val_opt unit_opt -> 
    let i =  
      match val_opt with
	  None -> assert false
	| Some id -> 
	    try
	      int_of_string (Ident.to_string id)
	    with
		_ -> assert false
    in
      Int(i, unit_opt)

let (aadl_ident_opt_to_addlreal : Ident.t option -> Ident.t option -> 
      aadl_real_or_constant) =
  fun val_opt unit_opt -> 
    let r =  
      match val_opt with
	  None -> assert false
	| Some id -> 
	    try
	      float_of_string (Ident.to_string id)
	    with
		_ -> assert false
    in
      Real(r, unit_opt)


let (aadl_ident_opt_to_int : Ident.t option -> int option) =
  fun val_opt  -> 
    match val_opt with
	None -> None
	| Some id ->  
	    try
	      Some (int_of_string (Ident.to_string id))
	    with
		_ -> assert false

let (aadl_ident_opt_to_real : Ident.t option -> float option) =
  fun val_opt  -> 
    match val_opt with
	None -> None
	| Some id ->  
	    try
	      Some (float_of_string (Ident.to_string id))
	    with
		_ -> assert false


let (aadl_ident_opt_to_string : Ident.t option -> string) = function
    None -> assert false
  | Some x  -> Ident.to_string x

let (aadl_ident_opt_to_ident : Ident.t option -> Ident.t) = function
    None -> Ident "unknown_enum_value"
  | Some x  -> x



let (parse_minmax_int : {{ [(minimum maximum)?] }} -> 
      aadl_intger_or_constant * aadl_intger_or_constant) =
  function
    | {{ [] }} -> assert false
    | {{ [pv_min  pv_max] }} -> 
	(* XXX fields "delta" and "base" are ignored.  *)
	let unit_min = seq_string_to_ident_opt {{ pv_min.?unitLiteral }}
	and unit_max = seq_string_to_ident_opt {{ pv_max.?unitLiteral }}
	in
	  ((match pv_min with
	      {{ <minimum value=min ..>[] }} 
	      -> 
		Int (int_of_string (ocamlduce_to_string min), unit_min)
	   ),
	   (match pv_max with
		{{ <maximum value=max ..>[] }} 
		-> 
		  Int (int_of_string (ocamlduce_to_string max), unit_max)
	   )
	  )
let (parse_minmax_real : {{ [(minimum maximum)?] }} -> 
      aadl_real_or_constant * aadl_real_or_constant) =
  function
    | {{ [] }} -> assert false
    | {{ [pv_min  pv_max] }} -> 
	let unit_min = seq_string_to_ident_opt {{ pv_min.?unitLiteral }}
	and unit_max = seq_string_to_ident_opt {{ pv_max.?unitLiteral }}
	in
	  ((match pv_min with
	      {{ <minimum value=min ..>[] }} 
	      -> 
		Real (float_of_string (ocamlduce_to_string min), unit_min)
	   ),
	   (match pv_max with
		{{ <maximum value=max ..>[] }} 
		-> 
		  Real (float_of_string (ocamlduce_to_string max), unit_max)
	   )
	  )

let (parse_property_expression: {{ propertyExpression }} -> property_expression) =
fun pv -> 
  let {{ <propertyValue..> [pv_body::Any*] }} = pv in
  let pchoice = {{ pv.xsi:type }} in
  let value_opt = seq_string_to_ident_opt {{ pv.?value }} in
  let unitLiteral_opt = seq_string_to_ident_opt {{ pv.?unitLiteral }} in
  let enumLiteral_opt = seq_string_to_ident_opt {{ pv.?enumLiteral }} in
  let delta_opt = seq_string_to_ident_opt {{ pv.?delta }} in

    (* XXX unused yet *)
  let _base_opt = 
    seq_string_to_ident_opt {{ pv.?base }} in
  let ref_elt_opt = seq_string_to_ident_opt {{ pv.?referenceElement }} in

  let expr = 
    match pchoice with
      | {{ "property:StringValue" }} -> 
	  Prop_expr_immediate(
	    Imm_const (Imm_string (aadl_ident_opt_to_string value_opt)))

      | {{ "property:BooleanValue" }} -> 
	  Prop_expr_immediate(
	    Imm_const (Imm_bool (aadl_ident_opt_to_bool value_opt)))

      | {{ "property:RealValue" }} -> 
	  Prop_expr_immediate(
	    Imm_const (Imm_real (aadl_ident_opt_to_addlreal value_opt unitLiteral_opt)))

      | {{ "property:IntegerValue" }} -> 
	  Prop_expr_immediate(
	    Imm_const (Imm_int (aadl_ident_opt_to_addlint value_opt unitLiteral_opt)))
	    
      | {{ "property:ReferenceValue" }} -> 
	  Prop_expr_reference_term (aadl_ident_opt_to_ident ref_elt_opt) 
	    
      | {{ "property:EnumValue" }} ->  
	  Prop_expr_enumeration (aadl_ident_opt_to_ident enumLiteral_opt) 
	    
      | {{ "property:IntegerRangeValue" }} ->
	  let (minimum, maximum) = parse_minmax_int pv_body in
	    Prop_expr_integer_range(minimum, maximum, 
				    aadl_ident_opt_to_int delta_opt)
	      
      | {{ "property:RealRangeValue" }} ->
	  let (minimum, maximum) = parse_minmax_real pv_body in
	    Prop_expr_real_range(minimum, maximum, aadl_ident_opt_to_real delta_opt)
  in
    expr
      
      
let (parse_properties: {{ properties }} -> property list) =
  fun f ->
    let {{ <properties> [body_seq::propertyAssociation+] }} = f in
    let (parse_propertyAssociation : {{ propertyAssociation }} -> property) =
      fun pa ->
	let {{ <propertyAssociation propertyDefinition=pdef ..> 
		 [pe::propertyExpression*] }} = pa
	in
	let applies_to_opt = seq_string_to_ident_opt {{ pa.?appliesTo }} in
	let prop_value_list = List.map parse_property_expression {: pe :} in
	  (
	    string_to_ident pdef,
	    Prop_arrow, (* "=>" is value by default ; 
			   "+=>" should be add somewhere in aadl_dtd.ml *)
	    prop_value_list,
	    applies_to_opt
	  )
    in
      List.map parse_propertyAssociation {: body_seq :}

(**********************************************************************************)
let (direction_to_in_out_flag : {{ [String?] }} -> in_out_flag) = function
  | {{ [] }} -> In (* It seems that no direction stands for In !!! *)
  |  {{ [str] }} -> 
       match ocamlduce_to_string str with 
	   "out" -> Out
	 | "in out" -> In_Out
	 | "in" -> In
	 | "" -> In (* It seems that no direction stands for In !!! *)
	 | "required" -> assert false
	 | "provided" -> assert false
	 | _  -> assert false

let (direction_to_access_kind : {{ [String?] }} -> access_kind) = function
  | {{ [] }} -> Provided (* It seems that no direction stands for Provided !!! *)
  |  {{ [str] }} -> 
       match ocamlduce_to_string str with 
	   "required" -> Required
	 | "provided" -> Provided
	 | "" -> Provided
	 | _ -> 
	     print_string (Ocamlduce.to_string str);
	     assert false

let (parse_features: {{ features }} -> feature list) =
  fun f ->
    let {{ <features> [body_seq::feature_body*] }} = f in
    let (parse_feature : {{ feature_body }} -> feature) =
      fun body -> 
      match body  with
	| {{ <busAccess name=name..>[p::properties*] }} ->
	    string_to_ident name,
	    Feat_subcomponent_access(
	      direction_to_access_kind {{ body.?direction }},
	      Bus,
	      seq_string_to_long_ident_opt {{ body.?busClassifier }}
	    ),
	    List.flatten (List.map parse_properties {: p :})
	      
	| {{ <dataAccess name=name..>[p::properties*] }} -> 
	    string_to_ident name,
	    Feat_subcomponent_access(
	      direction_to_access_kind {{ body.?direction }},
	      Data,
	      seq_string_to_long_ident_opt {{ body.?dataClassifier }}
	    ),
	    List.flatten (List.map parse_properties {: p :})
	      
	| {{ <dataPort name=name..>[p::properties*] }} ->
	    string_to_ident name,
	    Feat_data_port(
	      direction_to_in_out_flag {{ body.?direction }},
	      seq_string_to_long_ident_opt  {{ body.?dataClassifier }}
	    ),
	    List.flatten (List.map parse_properties {: p :})
	      
	| {{ <eventPort name=name..>[p::properties*] }} ->
	    string_to_ident name,
	    Feat_event_port(
	      direction_to_in_out_flag {{ body.?direction }}),
	    List.flatten (List.map parse_properties {: p :})
	      
	| {{ <eventDataPort name=name..>[p::properties*] }} ->
	    string_to_ident name,
	    Feat_event_data_port(
	      direction_to_in_out_flag {{ body.?direction }},
	      None
	    ),
	    List.flatten (List.map parse_properties {: p :})
	      
	| {{ <portGroup portGroupType=_pg_type name=name..>[p::properties*] }} ->
	    string_to_ident name,
	    Feat_port_group (
	      seq_string_to_long_ident_opt {{ body.?portGroupType }}
	    ),
	    List.flatten (List.map parse_properties {: p :})
	      
	| {{ <parameter name=name..>[p::properties*] }} ->
	    string_to_ident name,
	    Feat_parameter(
	      direction_to_in_out_flag {{ body.?direction }},
	      seq_string_to_long_ident_opt {{ body.?dataClassifier }}
	    ),
	    List.flatten (List.map parse_properties {: p :})
    in
      List.map parse_feature {: body_seq :}



(**********************************************************************************)
let (parse_flowSpecs: {{ flowSpecs }} -> flow list) =
  fun f ->
    (* flows are ignored *)
    []


(**********************************************************************************)
let (parse_flow: {{ flows }} -> flow list) =
  fun f ->
    (* flows are ignored *)
    []


(**********************************************************************************)
let (parse_callSequences : {{ callSequences }} -> call_sequence list) =
  fun cs -> 

    let {{<callSequences>[ call_seq::callSequence* ]}} = cs in 

    let (parse_call_sequence : {{ callSequence }} -> call_sequence ) =
      fun cs -> 
	let (parse_call : {{ call }} -> Ident.t * Ident.t * property_association list) =
	  fun c -> 
	    let {{<call ..>[prop::properties*]}} = c in
	      (
		string_to_ident {{ c.name }}, 
		string_to_long_ident {{ c.classifier }},
		List.flatten (List.map parse_properties {: prop :})
	      )
	in
	let {{<callSequence>[ cseq::call* ]}} = cs in
	  (
	    None, (* XXX je met quoi la ??? *)
	    List.map parse_call {: cseq :}
	  )
    in
      List.map parse_call_sequence {: call_seq :}

(**********************************************************************************)
let (parse_subcomponents : {{ subcomponents }} -> subcomponent list) =
  fun c -> 
    let {{ <subcomponents> [body_seq::subcomponents_body*] }} = c in
    let (parse_subcomponents_body : {{ subcomponents_body }} -> subcomponent) = 
      fun body  -> 
	let category, prop =
	  match body with
	    | {{<processSubcomponent ..>[prop::properties*]}} -> Process, prop
	    | {{<processorSubcomponent ..>[prop::properties*]}} -> Processor, prop
	    | {{<busSubcomponent ..>[prop::properties*]}} -> Bus, prop
	    | {{<dataSubcomponent ..>[prop::properties*]}} -> Data, prop
	    | {{<memorySubcomponent ..>[prop::properties*]}} -> Memory, prop
	    | {{<deviceSubcomponent ..>[prop::properties*]}} -> Device, prop
	    | {{<threadSubcomponent ..>[prop::properties*]}} -> Thread, prop
	in
	  (
	    string_to_ident {{ body.name }}, 
	    category,
	    seq_string_to_long_ident_opt {{ body.?classifier }},
	    List.flatten (List.map parse_properties {: prop :})
	  )
    in
      List.map parse_subcomponents_body {: body_seq :}

(**********************************************************************************)

let (string_to_connection_arrow :  {{  [String?] }} -> connection_arrow) = function
  | {{ [] }} -> Conn_arrow (* The default *)
  | {{ [str] }} ->
      match ocamlduce_to_string str with 
	| "Delayed" ->  Conn_delayed_arrow
	| "Immediate" -> Conn_arrow
	| _ -> failwith ("*** aaxl error: " ^ (Ocamlduce.to_string str) ^ 
			   " instead of Delayed or Immediate.\n")

(**********************************************************************************)
let (parse_connections : {{ connections }} -> connection list) =
  fun c -> 
    let {{ <connections> [body_seq::connections_body*] }} = c in
    let (parse_connection_body : {{ connections_body }} -> connection) = 
      fun body  -> 
	let connection_kind, prop =
	match body with
	  | {{<dataConnection..> [p::properties*]}}  -> Conn_data_port, p
	  | {{<eventConnection..> [p::properties*]}} -> Conn_event_port, p
	  | {{<eventDataConnection..> [p::properties*]}} -> Conn_event_data_port, p
	  | {{<parameterConnection..> [p::properties*]}} -> Conn_parameter, p
	  | {{<portGroupConnection..> [p::properties*]}} -> Conn_port_group, p
	  | {{<busAccessConnection..> [p::properties*]}} -> Conn_bus_access,p
	  | {{<dataAccessConnection..> [p::properties*]}} -> Conn_data_access, p
	in
	  (
	    seq_string_to_ident_opt {{ body.?name }},
	    connection_kind,
	    access_connection_string_to_ident
(* 	      (if  *)
(* 		 connection_kind = Conn_data_access *)
(* 	       then Some "_out"  *)
(* 	  else *)
	    None
(* 	      )  *)
	      {{ body.?srcContext }} 
	      {{ body.src }}, 
	    string_to_connection_arrow {{ body.?timing }}, 
	    access_connection_string_to_ident
(* 	      (if  *)
(* 		 connection_kind = Conn_data_access  *)
(* 	       then Some "_in"  *)
(* 	       else  *)
		 None
(* 	      ) *)
	      {{ body.?dstContext }} 
	      {{ body.dst }},
	    List.flatten (List.map parse_properties {: prop :})
	  )
    in
    let res =  List.map parse_connection_body {: body_seq :} in      
(*       List.iter *)
(* 	(fun (_,_,src,_,dst,_) ->  *)
(* 	   print_string (Ident.to_string src); *)
(* 	   print_string "  -> "; *)
(* 	   print_string (Ident.to_string dst); *)
(* 	   print_string "\n"; *)
(* 	) *)
      res

(**********************************************************************************)
let (from_component_type_body : 
       {{ [component_body_type*] }} -> (feature list * property list * flow list)) =
  fun data_body -> 
    List.flatten (List.map parse_features   {: data_body.(features) :}),
    List.flatten (List.map parse_properties {: data_body.(properties) :}),
    List.flatten (List.map parse_flowSpecs  {: data_body.(flowSpecs) :})


let (from_component_impl_body : {{ [component_body_impl*] }} -> 
      (subcomponent list *  call_sequence list * connection list * property list * 
	 flow list)) =
  fun data_body -> 
    List.flatten (List.map parse_subcomponents {: data_body.(subcomponents) :}),
    List.flatten (List.map parse_callSequences {: data_body.(callSequences) :}),
    List.flatten (List.map parse_connections   {: data_body.(connections) :}),
    List.flatten (List.map parse_properties    {: data_body.(properties) :}),
    List.flatten (List.map parse_flowSpecs     {: data_body.(flowSpecs) :})


(**********************************************************************************)
let (parse_threadType: {{ threadType }} -> aadl_declaration) =
  fun dt ->
    match dt with
	{{ <threadType name=str>[ data_body::component_body_type* ] }} ->
	  let (feature_l, properties_l, flowSpec_l) = 
	    from_component_type_body data_body 
	  in
	  let comp =
	    {
	      comp_id=aaxl_string_to_ident str; 
	      comp_kind=Type;
	      comp_cate=Thread;
	      comp_extd=None;
	      comp_feat=feature_l;
	      comp_fref=[];
	      comp_subc=[];
	      comp_call=[];
	      comp_conn=[];
	      comp_flow=flowSpec_l;
	      comp_mode=[];
	      comp_prop=properties_l; 
	      comp_annx=[]; 
	      comp_normalize=false;
	    }
	  in
	    Component comp


let (parse_threadImpl: {{ threadImpl }} -> aadl_declaration) =
  fun dt ->
    (* XXX The compType field provides the name the Type corresponding to the 
       implementation, but that name can be obtained from the name of the 
       component impl (cf p 37 of the standard). Therefore we ignore it here.
    *)
    let {{ <threadImpl name=str compType=_comp_type>
	     [ thread_body::component_body_impl* ] }} = dt in
    let (subcomponent_l, call_l, connections_l, properties_l, flow_l) = 
      from_component_impl_body thread_body in
    let comp =
      {
	comp_id=aaxl_string_to_ident str; 
	comp_kind=Implementation;
	comp_cate=Thread;
	comp_extd=None; 
	comp_feat=[];
	comp_fref=[];
	comp_subc=subcomponent_l;
	comp_call=call_l;
	comp_conn=connections_l;
	comp_flow=flow_l;
	comp_mode=[];
	comp_prop=properties_l; 
	comp_annx=[]; 
	comp_normalize=false;
      }
    in
      Component comp

(**********************************************************************************)

let (parse_busType: {{ busType }} -> aadl_declaration) =
  fun dt ->
    match dt with
	{{ <busType name=str>[ bus_body::component_body_type* ] }} ->
	  let (feature_l, properties_l, flowSpec_l) = from_component_type_body bus_body in
	  let comp =
	    {
	      comp_id=string_to_ident str; 
	      comp_kind=Type;
	      comp_cate=Bus;
	      comp_extd=None; 
	      comp_feat=feature_l;
	      comp_fref=[];
	      comp_subc=[];
	      comp_call=[];
	      comp_conn=[];
	      comp_flow=flowSpec_l;
	      comp_mode=[];
	      comp_prop=properties_l; 
	      comp_annx=[]; 
	      comp_normalize=false;
	    }
	  in
	    Component comp

let (parse_busImpl: {{ busImpl }} -> aadl_declaration) =
  fun dt ->
    let {{ <busImpl name=str compType=_comp_type>
	     [ bus_body::component_body_impl* ] }} = dt in
    let (subcomponent_l, call_l, connections_l, properties_l, flow_l) = 
      from_component_impl_body bus_body in
    let comp =
      {
	comp_id=string_to_ident str; 
	comp_kind=Implementation;
	comp_cate=Bus;
	comp_extd=None; 
	comp_feat=[];
	comp_fref=[];
	comp_subc=subcomponent_l;
	comp_call=call_l;
	comp_conn=connections_l;
	comp_flow=flow_l;
	comp_mode=[];
	comp_prop=properties_l; 
	comp_annx=[]; 
	comp_normalize=false;
      }
    in
      Component comp

(**********************************************************************************)

let (parse_memoryType: {{ memoryType }} -> aadl_declaration) =
  fun dt ->
    match dt with
	{{ <memoryType name=str>[ memory_body::component_body_type* ] }} ->
	  let (feature_l, properties_l, flowSpec_l) = from_component_type_body memory_body in
	  let comp =
	    {
	      comp_id=string_to_ident str; 
	      comp_kind=Type;
	      comp_cate=Memory;
	      comp_extd=None; 
	      comp_feat=feature_l;
	      comp_fref=[];
	      comp_subc=[];
	      comp_call=[];
	      comp_conn=[];
	      comp_flow=flowSpec_l;
	      comp_mode=[];
	      comp_prop=properties_l; 
	      comp_annx=[]; 
	      comp_normalize=false;
	    }
	  in
	    Component comp

let (parse_memoryImpl: {{ memoryImpl }} -> aadl_declaration) =
  fun dt ->
    let {{ <memoryImpl name=str compType=_comp_type>
	     [ memory_body::component_body_impl* ] }} = dt in
    let (subcomponent_l, call_l, connections_l, properties_l, flow_l) = 
      from_component_impl_body memory_body in
    let comp =
      {
	comp_id=string_to_ident str; 
	comp_kind=Implementation;
	comp_cate=Memory;
	comp_extd=None; 
	comp_feat=[];
	comp_fref=[];
	comp_subc=subcomponent_l;
	comp_call=call_l;
	comp_conn=connections_l;
	comp_flow=flow_l;
	comp_mode=[];
	comp_prop=properties_l; 
	comp_annx=[]; 
	comp_normalize=false;
      }
    in
      Component comp

(**********************************************************************************)
let (parse_processType: {{ processType }} -> aadl_declaration) =
  fun dt ->
    match dt with
	{{ <processType name=str>[ process_body::component_body_type* ] }} ->
	  let (feature_l, properties_l, flowSpec_l) = 
	    from_component_type_body process_body 
	  in
	  let comp =
	    {
	      comp_id=string_to_ident str; 
	      comp_kind=Type;
	      comp_cate=Process;
	      comp_extd=None; 
	      comp_feat=feature_l;
	      comp_fref=[];
	      comp_subc=[];
	      comp_call=[];
	      comp_conn=[];
	      comp_flow=flowSpec_l;
	      comp_mode=[];
	      comp_prop=properties_l; 
	      comp_annx=[]; 
	      comp_normalize=false;
	    }
	  in
	    Component comp

let (parse_processImpl: {{ processImpl }} -> aadl_declaration) =
  fun dt ->
    let {{ <processImpl name=str compType=_comp_type>
	     [ process_body::component_body_impl* ] }} = dt in
    let (subcomponent_l, call_l, connections_l, properties_l, flow_l) = 
      from_component_impl_body process_body in
    let comp =
      {
	comp_id=string_to_ident str; 
	comp_kind=Implementation;
	comp_cate=Process;
	comp_extd=None; 
	comp_feat=[];
	comp_fref=[];
	comp_subc=subcomponent_l;
	comp_call=call_l;
	comp_conn=connections_l;
	comp_flow=flow_l;
	comp_mode=[];
	comp_prop=properties_l; 
	comp_annx=[]; 
	comp_normalize=false;
      }
    in
      Component comp

(**********************************************************************************)
let (parse_processorType: {{ processorType }} -> aadl_declaration) =
  fun dt ->
    match dt with
	{{ <processorType name=str>[ processor_body::component_body_type* ] }} ->
	  let (feature_l, properties_l, flowSpec_l) = from_component_type_body processor_body in
	  let comp =
	    {
	      comp_id=string_to_ident str; 
	      comp_kind=Type;
	      comp_cate=Processor;
	      comp_extd=None; 
	      comp_feat=feature_l;
	      comp_fref=[];
	      comp_subc=[];
	      comp_call=[];
	      comp_conn=[];
	      comp_flow=flowSpec_l;
	      comp_mode=[];
	      comp_prop=properties_l; 
	      comp_annx=[]; 
	      comp_normalize=false;
	    }
	  in
	    Component comp

let (parse_processorImpl: {{ processorImpl }} -> aadl_declaration) =
  fun dt ->
    let {{ <processorImpl name=str compType=_comp_type>
	     [ processor_body::component_body_impl* ] }} = dt in
    let (subcomponent_l, call_l, connections_l, properties_l, flow_l) = 
      from_component_impl_body processor_body in
    let comp =
      {
	comp_id=string_to_ident str; 
	comp_kind=Implementation;
	comp_cate=Processor;
	comp_extd=None; 
	comp_feat=[];
	comp_fref=[];
	comp_subc=subcomponent_l;
	comp_call=call_l;
	comp_conn=connections_l;
	comp_flow=flow_l;
	comp_mode=[];
	comp_prop=properties_l; 
	comp_annx=[]; 
	comp_normalize=false;
      }
    in
      Component comp

(**********************************************************************************)
let (parse_deviceType: {{ deviceType }} -> aadl_declaration) =
  fun dt ->
    match dt with
	{{ <deviceType name=str>[ device_body::component_body_type* ] }} ->
	  let (feature_l, properties_l, flowSpec_l) = from_component_type_body device_body in
	  let comp =
	    {
	      comp_id=string_to_ident str; 
	      comp_kind=Type;
	      comp_cate=Device;
	      comp_extd=None; 
	      comp_feat=feature_l;
	      comp_fref=[];
	      comp_subc=[];
	      comp_call=[];
	      comp_conn=[];
	      comp_flow=flowSpec_l;
	      comp_mode=[];
	      comp_prop=properties_l; 
	      comp_annx=[]; 
	      comp_normalize=false;
	    }
	  in
	    Component comp

let (parse_deviceImpl: {{ deviceImpl }} -> aadl_declaration) =
  fun dt ->
    let {{ <deviceImpl name=str compType=_comp_type>
	     [ device_body::component_body_impl* ] }} = dt in
    let (subcomponent_l, call_l, connections_l, properties_l, flow_l) = 
      from_component_impl_body device_body in
    let comp =
      {
	comp_id=string_to_ident str; 
	comp_kind=Implementation;
	comp_cate=Device;
	comp_extd=None; 
	comp_feat=[];
	comp_fref=[];
	comp_subc=subcomponent_l;
	comp_call=call_l;
	comp_conn=connections_l;
	comp_flow=flow_l;
	comp_mode=[];
	comp_prop=properties_l; 
	comp_annx=[]; 
	comp_normalize=false;
      }
    in
      Component comp

(**********************************************************************************)
let (parse_threadType: {{ threadType }} -> aadl_declaration) =
  fun dt ->
    match dt with
	{{ <threadType name=str>[ thread_body::component_body_type* ] }} ->
	  let (feature_l, properties_l, flowSpec_l) = 
	    from_component_type_body thread_body 
	  in
	  let comp =
	    {
	      comp_id=string_to_ident str; 
	      comp_kind=Type;
	      comp_cate=Thread;
	      comp_extd=None; 
	      comp_feat=feature_l;
	      comp_fref=[];
	      comp_subc=[];
	      comp_call=[];
	      comp_conn=[];
	      comp_flow=flowSpec_l;
	      comp_mode=[];
	      comp_prop=properties_l; 
	      comp_annx=[]; 
	      comp_normalize=false;
	    }
	  in
 	    Component comp

let (parse_threadImpl: {{ threadImpl }} -> aadl_declaration) =
  fun dt ->
    let {{ <threadImpl name=str compType=_comp_type>
	     [ thread_body::component_body_impl* ] }} = dt in
    let (subcomponent_l, call_l, connections_l, properties_l, flow_l) = 
      from_component_impl_body thread_body in
    let comp =
      {
	comp_id=string_to_ident str; 
	comp_kind=Implementation;
	comp_cate=Thread;
	comp_extd=None; 
	comp_feat=[];
	comp_fref=[];
	comp_subc=subcomponent_l;
	comp_call=call_l;
	comp_conn=connections_l;
	comp_flow=flow_l;
	comp_mode=[];
	comp_prop=properties_l; 
	comp_annx=[]; 
	comp_normalize=false;
      }
    in
      Component comp


(**********************************************************************************)
let (parse_systemType: {{ systemType }} -> aadl_declaration) =
  fun dt ->
    match dt with
	{{ <systemType name=str>[ system_body::component_body_type* ] }} ->
	  let (feature_l, properties_l, flowSpec_l) = from_component_type_body system_body in
	  let comp =
	    {
	      comp_id=string_to_ident str; 
	      comp_kind=Type;
	      comp_cate=System;
	      comp_extd=None; 
	      comp_feat=feature_l;
	      comp_fref=[];
	      comp_subc=[];
	      comp_call=[];
	      comp_conn=[];
	      comp_flow=flowSpec_l;
	      comp_mode=[];
	      comp_prop=properties_l; 
	      comp_annx=[]; 
	      comp_normalize=false;
	    }
	  in
	    Component comp

let (parse_systemImpl: {{ systemImpl }} -> aadl_declaration) =
  fun dt ->
    let {{ <systemImpl name=str compType=_comp_type>
	     [ system_body::component_body_impl* ] }} = dt in
    let (subcomponent_l, call_l, connections_l, properties_l, flow_l) = 
      from_component_impl_body system_body in
    let comp =
      {
	comp_id=string_to_ident str; 
	comp_kind=Implementation;
	comp_cate=System;
	comp_extd=None; 
	comp_feat=[];
	comp_fref=[];
	comp_subc=subcomponent_l;
	comp_call=call_l;
	comp_conn=connections_l;
	comp_flow=flow_l;
	comp_mode=[];
	comp_prop=properties_l; 
	comp_annx=[]; 
	comp_normalize=false;
      }
    in
      Component comp

(**********************************************************************************)
let (parse_subprogramType: {{ subprogramType }} -> aadl_declaration) =
  fun dt ->
    match dt with
	{{ <subprogramType name=str>[ subprogram_body::component_body_type* ] }} ->
	  let (feature_l, properties_l, flowSpec_l) = from_component_type_body subprogram_body in
	  let comp =
	    {
	      comp_id=string_to_ident str; 
	      comp_kind=Type;
	      comp_cate=Subprogram;
	      comp_extd=None; 
	      comp_feat=feature_l;
	      comp_fref=[];
	      comp_subc=[];
	      comp_call=[];
	      comp_conn=[];
	      comp_flow=flowSpec_l;
	      comp_mode=[];
	      comp_prop=properties_l; 
	      comp_annx=[]; 
	      comp_normalize=false;
	    }
	  in
	    Component comp

let (parse_subprogramImpl: {{ subprogramImpl }} -> aadl_declaration) =
  fun dt ->
    let {{ <subprogramImpl name=str compType=_comp_type>
	     [ subprogram_body::component_body_impl* ] }} = dt in
    let (subcomponent_l, call_l, connections_l, properties_l, flow_l) = 
      from_component_impl_body subprogram_body in
    let comp =
      {
	comp_id=string_to_ident str; 
	comp_kind=Implementation;
	comp_cate=Subprogram;
	comp_extd=None; 
	comp_feat=[];
	comp_fref=[];
	comp_subc=subcomponent_l;
	comp_call=call_l;
	comp_conn=connections_l;
	comp_flow=flow_l;
	comp_mode=[];
	comp_prop=properties_l; 
	comp_annx=[]; 
	comp_normalize=false;
      }
    in
      Component comp

(**********************************************************************************)

let (parse_dataType: {{ dataType }} -> aadl_declaration) =
  fun dt ->
    match dt with
	{{ <dataType name=str>[ data_body::component_body_type* ] }} ->
	  let (feature_l, properties_l, flowSpec_l) = from_component_type_body data_body in
	  let comp =
	    {
	      comp_id=string_to_ident str; 
	      comp_kind=Type;
	      comp_cate=Data;
	      comp_extd=None; 
	      comp_feat=feature_l;
	      comp_fref=[];
	      comp_subc=[];
	      comp_call=[];
	      comp_conn=[];
	      comp_flow=flowSpec_l;
	      comp_mode=[];
	      comp_prop=properties_l; 
	      comp_annx=[]; 
	      comp_normalize=false;
	    }
	  in
	    Component comp


let (parse_dataImpl: {{ dataImpl }} -> aadl_declaration) =
  fun dt ->
    let {{ <dataImpl name=str compType=_comp_type>
	     [ data_body::component_body_impl* ] }} = dt in
    let (subcomponent_l, call_l, connections_l, properties_l, flow_l) = 
      from_component_impl_body data_body in
    let comp =
      {
	comp_id=string_to_ident str; 
	comp_kind=Implementation;
	comp_cate=Data;
	comp_extd=None; 
	comp_feat=[];
	comp_fref=[];
	comp_subc=subcomponent_l;
	comp_call=call_l;
	comp_conn=connections_l;
	comp_flow=flow_l;
	comp_mode=[];
	comp_prop=properties_l; 
	comp_annx=[]; 
	comp_normalize=false;
      }
    in
      Component comp

(**********************************************************************************)
let (parse_portGroupType: {{ portGroupType }} -> aadl_declaration) =
  fun pg ->
    match pg with
	{{ <portGroupType name=n ..>[ comment_or_features::_* ] }} ->
	  let features = 
	    List.flatten 
	      (List.map parse_features {: comment_or_features.(features) :})
	  in
	  let features_ref = [] in (* XXX pas trait pour l'instant (li au raffinement)*)
	  let comp =
	    {
	      portgrp_id=string_to_ident n;
	      portgrp_extd=None; (* XXX *)
	      portgrp_feat=features;
	      portgrp_fref=features_ref; 
	      portgrp_inve= 
		(match {{ pg.?inverseOf }} with
		   | {{ [] }} -> None
		   | {{ [str]}} -> Some (string_to_ident str)
		); 
	      portgrp_prop=[]; 
	      portgrp_annx=[]; 
	      portgrp_normalize=false
	    }
	  in
	    Port_group comp


let (parse_component_declaration : 
       {{ component_declaration }} -> aadl_declaration) =
  fun pack -> 
    match pack with
      | {{ x & busImpl }} -> parse_busImpl x
      | {{ x & busType }} -> parse_busType x
      | {{ x & deviceType }} -> parse_deviceType x
      | {{ x & deviceImpl }} -> parse_deviceImpl x
      | {{ x & memoryImpl }} -> parse_memoryImpl x
      | {{ x & memoryType }} -> parse_memoryType x
      | {{ x & processImpl }} -> parse_processImpl x  
      | {{ x & processType }} -> parse_processType x 
      | {{ x & processorImpl }} -> parse_processorImpl x
      | {{ x & processorType }} -> parse_processorType x
      | {{ x & subprogramType }} -> parse_subprogramType x 
      | {{ x & subprogramImpl }} -> parse_subprogramImpl x 
      | {{ x & threadImpl }} -> parse_threadImpl x
      | {{ x & threadType }} -> parse_threadType x 
      | {{ x & dataType }} -> parse_dataType x
      | {{ x & dataImpl }} -> parse_dataImpl x
      | {{ x & portGroupType }} -> parse_portGroupType x
      | {{ x & systemType }} -> parse_systemType x
      | {{ x & systemImpl }} -> parse_systemImpl x


let (parse_aadlPublic : 
       {{ aadlPublic }} -> aadl_declaration list) =
  fun x -> 
    let {{ <aadlPublic> [ pack::component_declaration* ] }} = x in
      List.map parse_component_declaration {: pack :}

let (parse_aadlPrivate : 
       {{ aadlPrivate }} -> aadl_declaration list) =
  fun x -> 
    let {{ <aadlPrivate> [ pack::component_declaration* ] }} = x in
      List.map parse_component_declaration {: pack :}



let (parse_package : {{ String }} -> {{ aadlPackage }} -> Aadl_ast.package) =
    fun name_pack decl -> 
      match decl with
	  {{ x & aadlPublic }} -> 
	    {
	      pack_id = (string_to_ident name_pack);
	      pack_public = parse_aadlPublic x;
	      pack_private = [];
	      pack_public_prop = [];
	      pack_private_prop = [];
	    }
	| {{ x & aadlPrivate  }} ->
	    {
	      pack_id = (string_to_ident name_pack);
	      pack_public = [];
	      pack_private = parse_aadlPrivate x; 
	      pack_public_prop = [];
	      pack_private_prop = [];
	    }
	| {{ [x y] & [aadlPublic aadlPrivate]  }} -> 
	    {
	      pack_id = (string_to_ident name_pack);
	      pack_public = parse_aadlPublic x;
	      pack_private = parse_aadlPrivate y; 
	      pack_public_prop = [];
	      pack_private_prop = [];
	    }

let (parse_package_declaration: {{ aadlPackage_declaration }} -> Aadl_ast.package) =
  fun decl -> 
    match decl with
      | {{ <aadlPackage name=name_pack> [aadl_pack]  }} ->
	  parse_package name_pack aadl_pack  



let (parse_declarations: 
       Aadl_ast.package list * aadl_declaration list -> 
      {{ aadlDeclaration }} -> Aadl_ast.package list * aadl_declaration list) =
  fun (pl,dl) decl -> 
    match decl with
      | {{ x & aadlPackage_declaration }} -> (parse_package_declaration x)::pl, dl
      | {{ x & component_declaration }} -> pl, (parse_component_declaration x)::dl





let (parse_aadl : int -> string -> Aadl_ast.package list) =
  fun verb_level file ->
    let (aadl_spec : Aadl_dtd.aadlSpec) = load_aaxl verb_level file in
      (*
	 At this stage, we know that aaxl_file is well-type according to
	 the ocamlduce type defined in Aadl_dtd.

	 We're gonna transform it into a plain ocaml data-type.
      *)

      let {{ <core:AadlSpec..> [spec_body::aadlDeclaration*] }} = aadl_spec in
      let (pack_list, decl_list) =
	List.fold_left parse_declarations ([],[]) {: spec_body :}
      in
	if decl_list = [] then pack_list else
	  {
	    pack_id = Ident (Filename.chop_extension (Filename.basename file));
	    pack_public = decl_list;
	    pack_private = []; 
	    pack_public_prop = [];
	    pack_private_prop = [];
	  } :: pack_list
(* 	List.map parse_declarations {: spec_body :} *)

	
