(** Time-stamp: <modified the 18/01/2008 (at 15:04) by Erwan Jahier> *)
(*************************************************************************)
(*  AADL                                                                 *)
(*                                                                       *)
(*  File : global.ml                                                     *)
(*  Authors : Louis Mandel, Erwan Jahier                                 *)
(*  Organization : Verimag/CNRS                                          *)
(*                                                                       *)
(*************************************************************************)
(* $Id$ *)

(** 
    @author Louis Mandel (VERIMAG)
    @author Erwan Jahier (VERIMAG)
*)

type t = 
  | Ident of string 
  | Dot of string * t
  | Col of string * t

type integer = 
  | Id of t
  | I of int

type tok = STR of string | DOT | COL

let rec (to_string:  t -> string) = 
  fun id -> 
    let str = 
      match id with
	| Ident s -> s
	| Dot (s,id') -> s ^ "_"  ^ (to_string id')
	| Col (s,id') -> s ^ "__" ^ (to_string id')
    in
      str
   
let rec (integer_to_string:  integer -> string) =
  function
    | Id id -> to_string id
    | I i -> string_of_int i
	  
	
let rec (add_suffix:string -> t -> t) =
  fun suff id -> 
    match id with
    | Ident s -> Ident (s^suff)
    | Dot (s,id') -> Dot (s, add_suffix suff id')
    | Col (s,id') -> Col (s, add_suffix suff id')
  
(* exported *)
let (make_new_ident : t -> string -> t) =
  fun ident suffix -> 
    (Ident ("_" ^ (to_string ident) ^ suffix))

let (of_string : string -> t) =
  fun str -> 
    let rec insert_element elt l =
	match l with
	    [] -> []
	  | [x] -> [x]
	  | x::tail -> x::elt::(insert_element elt tail)
    in
    let str_l = Str.split (Str.regexp "::") str in
    let str_ll = List.map (Str.split (Str.regexp "\\.")) str_l in
    let tok_ll = List.map (List.map (fun str -> STR str)) str_ll in
    let tok_dot_ll =  List.map (insert_element DOT) tok_ll in
    let tok_dot_col_l = List.flatten (insert_element [COL] tok_dot_ll) in
      (* 
	 [tok_dot_col_l] contains lists that look like:
	     [STR "FOO"; COL; STR "BAR"; DOT ; STR "f"]
	 and turning it into a Ident.T is easy, look:
      *)
    let rec (tok_list_to_ident : tok list -> t) =
      function
	| [STR str] -> Ident str
	| (STR str)::COL::tail -> Col(str, tok_list_to_ident tail)
	| (STR str)::DOT::tail -> Dot(str, tok_list_to_ident tail)
	| _ -> assert false
    in
      tok_list_to_ident tok_dot_col_l

let _ =
  assert (of_string "f" = Ident "f");
  assert ( (of_string "FOO::BAR.f") = Col("FOO",Dot("BAR",Ident "f")))



(* Splits aaxl ident into a list of pairs (cf the assert below). *)
let (split_aaxl_ident : string -> (string * string) list )=
  fun str -> 
    let get_next_indexes i1 =
      let i2 = Str.search_forward (Str.regexp_string "[@name") str i1 in
      let i3 = Str.search_forward (Str.regexp_string "]") str i2 in
	(i2, i3)
    in
    let get_pair i1 i2 i3 =
      (String.sub str i1 (i2-i1)), (String.sub str  (i2+7) (i3-i2-7))
    in
    let rec loop acc i1 =
      try
	let (i2, i3) = get_next_indexes i1 in
	  loop ((get_pair i1 i2 i3)::acc) (i3+1)
      with _ -> 
	acc
    in
      loop ["",str] 0 


let _ = 
  let str = "/aadlSpec[@name=Test_process_impl]/aadlPackage[@name=Test_process_impl]/aadlPublic/processImpl[@name=P.IMPL]" in
  assert (
    (split_aaxl_ident str) =
      [("/aadlPublic/processImpl", "P.IMPL");
       ("/aadlPackage", "Test_process_impl"); 
       ("/aadlSpec", "Test_process_impl");
       ("",
	"/aadlSpec[@name=Test_process_impl]/aadlPackage[@name=Test_process_impl]/aadlPublic/processImpl[@name=P.IMPL]")]
  )


let (connection_string_to_ident : string -> string -> t) =
  fun ctx port ->
    let s1 = snd (List.hd (split_aaxl_ident ctx))
    and s2 = snd (List.hd (split_aaxl_ident port))
    in
      if
	(Str.string_match 
	   (Str.regexp "^/subcomponents") 
	   (fst (List.hd (split_aaxl_ident ctx))) 0)
	  or
	(Str.string_match 
	   (Str.regexp "^/call") 
	   (fst (List.hd (split_aaxl_ident ctx))) 0)
      then
	of_string (s1 ^ "." ^ s2)
      else
	(* This ougth to correspond to the case where the port is local 
	   (i.e., it is a port of the current component) *)
	of_string s2

let rec (is_prefix:  t -> t -> bool) =
  fun id1 id2 ->
    begin match id1 with
      | Ident n1 ->
	  begin match id2 with
	    | Ident n2 -> n1 = n2
	    | Dot (n2, id2') | Col (n2, id2') -> n1 = n2
	  end

      | Dot (n1, id1')  -> 
	  begin match id2 with
	    | Dot (n2, id2') -> 
		if n1 = n2 then
		  is_prefix id1' id2'
		else 
		  false
	    | _ -> false
	  end

      | Col (n1, id1')  -> 
	  begin match id2 with
	    | Col (n2, id2') -> 
		if n1 = n2 then
		  is_prefix id1' id2'
		else 
		  false
	    | _ -> false
	  end
	    
    end


let (aaxl_string_to_string: string -> string) = 
  fun str -> 
    try (snd (List.hd (split_aaxl_ident str)))
    with _ -> 
	failwith (
	  "*** cannot convert into an identifier the string \"" ^ str ^ "\"\n")

let (aaxl_string_to_ident : string -> t) = 
  fun str -> of_string  (aaxl_string_to_string str)

let (aaxl_string_to_short_ident:  string -> t) = aaxl_string_to_ident
  
(* this one is just a guess that seems to work... *)
let (aaxl_string_to_long_ident:  string -> t) =
  fun str ->
    let rec aux l =
      begin match l with
	| [] -> ""
	| (tag, name) :: l ->
	    if Str.string_match (Str.regexp "^/aadlPackage") tag 0 
	    then name else aux l 
      end
    in
    let l = split_aaxl_ident str in
    let id = aaxl_string_to_string str in
    let pack_name = aux (List.tl l) in
    let str' = pack_name ^ "::" ^ id in
(*       print_string ("\n\t *** aaxl_string_to_long_ident \"" ^ str ^ "\"\n\t\t = " ^ str'); *)
      of_string str'




let rec (type_name_of_implementation_name:  t -> t) =
  fun id ->
    begin match id with
      | Ident _ -> assert false
      | Dot (x, Ident _) -> Ident x
      | Dot (x, id') -> Dot(x, type_name_of_implementation_name id')
      | Col (x, id') -> Col(x, type_name_of_implementation_name id')
    end


let rec (feature_name_of_component_name:  t -> t) =
  fun id ->
    begin match id with
      | Ident _ -> id 
      | Dot (_, id') | Col (_, id') -> feature_name_of_component_name id'
    end

(** [make_add_prefix id] cre une fonction qui ajoute le nom [id] 
     un identifcateur
*)
let rec (make_add_prefix:  t -> t -> t) =
  fun comp_id ->
    begin match comp_id with
      | Col (s, id) -> 
	  let add_prefix = make_add_prefix id in
	    (fun x -> Col(s, add_prefix x))
      | Dot (s, id) -> 
	  let add_prefix = make_add_prefix id in
	    (fun x -> Dot(s, add_prefix x))
      | Ident s ->
	  (fun x -> Dot(s, x))
    end

      

